#  Copyright 2021 Google LLC
#  Copyright 2021 EPAM Systems
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import copy
import json
import re
import logging
from typing import List, re as regex

from libs.context import Context
from libs.exceptions import EmptyManifestError
from libs.search_record_ids import ExtendedSearchId

logger = logging.getLogger()


class ManifestIntegrity(object):
    """Class to validate if parents reference and master data are exists and
    remove non-valid entities to provide integrity
    """
    REFERENCE_DATA_ID_PATTERN = re.compile(r"[\w\-\.]+:reference-data\-\-[\w\-\.]+:.[^,;\"]+",
                                           re.I + re.M)
    MASTER_DATA_ID_PATTERN = re.compile(r"[\w\-\.]+:master-data\-\-[\w\-\.]+:.[^,;\"]+",
                                        re.I + re.M)
    WORK_PRODUCT_ID_PATTERN = re.compile(r"[\w\-\.]+:work-product\-\-[\w\-\.]+:.[^,;\"]+",
                                         re.I + re.M)
    WORK_PRODUCT_COMPONENT_ID_PATTERN = re.compile(
        r"[\w\-\.]+:work-product-component\-\-[\w\-\.]+:.[^,;\"]+", re.I + re.M)
    DATASET_ID_PATTERN = re.compile(r"[\w\-\.]+:dataset\-\-[\w\-\.]+:.[^,;\"]+", re.I + re.M)

    def __init__(self, search_url: str, token_refresher, context: Context):
        self.search_url = search_url
        self.token_refresher = token_refresher
        self.context = context
        self.ids_for_validation = []
        self.entities_ids = {}
        self.patterns = [self.REFERENCE_DATA_ID_PATTERN,
                         self.MASTER_DATA_ID_PATTERN,
                         self.WORK_PRODUCT_ID_PATTERN,
                         self.WORK_PRODUCT_COMPONENT_ID_PATTERN,
                         self.DATASET_ID_PATTERN]

    def _match_id_with_pattern(self, pattern: regex.Pattern, source: str) -> List[str]:
        return pattern.findall(source)

    def _collect_ids(self, manifest_file: dict):
        """
        Collect manifest entities ids to exclude them while integrity checking
        """
        entities_ids = []
        if manifest_file.get("ReferenceData"):
            for elem in manifest_file["ReferenceData"]:
                if elem.get("id"):
                    entities_ids.append(elem["id"])
        if manifest_file.get("MasterData"):
            for elem in manifest_file["MasterData"]:
                if elem.get("id"):
                    entities_ids.append(elem["id"])
        if manifest_file.get("Data"):
            if manifest_file["Data"].get("WorkProduct") and manifest_file["Data"]["WorkProduct"].get("id"):
                entities_ids.append(manifest_file["Data"]["WorkProduct"]["id"])
            if manifest_file["Data"].get("WorkProductComponents"):
                for elem in manifest_file["Data"]["WorkProductComponents"]:
                    if elem.get("id"):
                        entities_ids.append(elem["id"])
            if manifest_file["Data"].get("Datasets"):
                for elem in manifest_file["Data"]["Datasets"]:
                    if elem.get("id"):
                        entities_ids.append(elem["id"])
        self.entities_ids = set(entities_ids)

    def _extract_references(self, manifest: dict) -> List[str]:
        """
        Extract reference ids from manifest
        """
        temp_manifest_entity = copy.deepcopy(manifest)
        temp_manifest_entity.pop("kind")
        manifest_str = json.dumps(temp_manifest_entity)
        ids_for_validation = []
        for pattern in self.patterns:
            ids_for_validation.extend(self._match_id_with_pattern(pattern, manifest_str))
        ids_for_validation = list(set(ids_for_validation).difference(self.entities_ids))
        logger.debug(ids_for_validation)
        return ids_for_validation

    def _validate_entity(self, manifest: dict) -> bool:
        """
        Validate reference ids from manifest entity
        """
        ids_for_validation = self._extract_references(manifest)
        missing_ids = None
        if ids_for_validation:
            search_class = ExtendedSearchId(self.search_url, ids_for_validation, self.token_refresher, self.context)
            found_ids, missing_ids = search_class.search_records()
        if not missing_ids:
            return True
        return False

    def ensure_integrity(self, manifest_file: dict=None):
        """
        Validate reference ids in traversal manifest file

        :return: only valid entities in the same structure
        """

        if not manifest_file:
            raise EmptyManifestError()

        self._collect_ids(manifest_file)

        if manifest_file.get("ReferenceData"):
            for data in manifest_file["ReferenceData"]:
                is_valid = self._validate_entity(data)
                if not is_valid:
                    logger.warning(f"Resource with kind {data.get('kind')} was rejected")
                    manifest_file["ReferenceData"].remove(data)
        if manifest_file.get("MasterData"):
            master_data = manifest_file["MasterData"]
            for data in master_data:
                is_valid = self._validate_entity(data)
                if not is_valid:
                    logger.warning(f"Resource with kind {data.get('kind')} was rejected")
                    manifest_file["MasterData"].remove(data)
        if manifest_file.get("Data"):
            if manifest_file["Data"].get("WorkProduct"):
                work_product_data = manifest_file["Data"]["WorkProduct"]
                is_valid = self._validate_entity(work_product_data)
                if not is_valid:
                    logger.warning(f"Resource with kind {work_product_data.get('kind')} was rejected")
                    manifest_file["Data"]["WorkProduct"] = {}
            if manifest_file["Data"].get("WorkProductComponents"):
                wpc_data = manifest_file["Data"]["WorkProductComponents"]
                for data in wpc_data:
                    is_valid = self._validate_entity(data)
                    if not is_valid:
                        logger.warning(f"Resource with kind {data.get('kind')} was rejected")
                        manifest_file["Data"]["WorkProductComponents"].remove(data)
            if manifest_file["Data"].get("Datasets"):
                datasets = manifest_file["Data"]["Datasets"]
                for data in datasets:
                    is_valid = self._validate_entity(data)
                    if not is_valid:
                        logger.warning(f"Resource with kind {data.get('kind')} was rejected")
                        manifest_file["Data"]["Datasets"].remove(data)
